// -------------------------------------------------------
//
// Author: Thomas Lenart    
// Email:  thomas.lenart@es.lth.se
// Date:   March 2003
//
// -------------------------------------------------------

#include "main.h"
#include "memhandler.h"
#include "buffer.h"
#include "decoder.h"


CT_Buffer *buffer_create(void)
{
	CT_Buffer *buffer;

	buffer = (CT_Buffer*)malloc(sizeof(CT_Buffer));			// allocate a new buffer

	buffer->first = 0;										// clear first pointer
	buffer->last  = 0;										// clear last pointer

	return buffer;
}

CT_Element *buffer_add(CT_Buffer *buffer, unsigned char *data, int data_size, void *header, int header_type)
{
	CT_Element *element;

	element = (CT_Element*)malloc(sizeof(CT_Element));		// allocate a new element

	element->data = data;									// set data pointer
	element->data_size = data_size;							// set data size
	element->header = header;								// set data pointer
	element->header_type = header_type;						// set data pointer
	element->next = 0;										// clear next pointer
	if (buffer->last)										// if not empty
		buffer->last->next = element;						// point to this element
	else
		buffer->first = element;							// update first pointer
	buffer->last = element;									// update last pointer
		
	return element;
}

int buffer_delete(CT_Buffer *buffer, int options)
{
	CT_Element *element;

	if (!buffer->first)										// if buffer empty
		return CT_BUFFER_ERROR;								// return error

	element = buffer->first;								// save the first element
	buffer->first = buffer->first->next;					// move the pointer to the second element
	if ((element->header) && (options & CT_DELETE_HEADER))
		free(element->header);								// free data pointer
	if ((element->data) && (options & CT_DELETE_DATA))
		free(element->data);								// free data pointer
	free(element);	// free element pointer

	if (!buffer->first)										// if buffer empty
		buffer->last = 0;									// clear last pointer

	return CT_BUFFER_OK;
}

void buffer_destroy(CT_Buffer *buffer)
{
	CT_Element *element = buffer->first;					// get the first element
	
	while (element) {										// while not empty
		element = element->next;							// save next pointer
		buffer_delete(buffer, CT_DELETE_ALL);
	}

	free(buffer);
}

int buffer_empty(CT_Buffer *buffer)
{
	if (buffer->first)										// if buffer is not empty
		return 0;											// return false
	
	return 1;												// else return true
}

int buffer_size(CT_Buffer *buffer)
{
	CT_Element	*element = buffer->first;					// get the first element
	int			total_size = 0;								// total buffer size

	while (element) {										// while not empty
		total_size += element->data_size;
		element = element->next;							// get next element
	}
	
	return total_size;
}

int buffer_count(CT_Buffer *buffer)
{
	CT_Element	*element = buffer->first;					// get the first element
	int			size = 0;									// total buffer size

	while (element) {										// while not empty
		size++;
		element = element->next;							// get next element
	}
	
	return size;
}

int buffer_move(CT_Buffer *from, CT_Buffer *to)
{
	if (!from->first)										// return if empty
		return 0;

	buffer_add(to, from->first->data, from->first->data_size,
				   from->first->header, from->first->header_type);	// insert element in destination
	return 1;
}

int buffer_move_all(CT_Buffer *from, CT_Buffer *to)
{
	int items = 0;

	while (buffer_move(from, to)) { items++; }				// move all elements

	if (items)												// if elements moved
		return 1;											// return true
	return 0;												// else return false
}	

/*
void buffer_debug_info(char* name, CT_Buffer *buffer)
{
	int i;
	CT_Element	*element = buffer->first;		

	printf("\n%s\n---------------\n", name);
	printf("Elements : %d\n", buffer_count(buffer));

	for (i = 0; i < buffer_count(buffer); i++) {
		printf("%d : data = %p size = %d\n", i, element->data, element->data_size);
		element = element->next;
	}
}
*/
