#ifndef __BUFFER_H__
#define __BUFFER_H__

#include "constants.h"

#define CT_BUFFER_OK		1
#define CT_BUFFER_ERROR		0

struct _CT_Element {
	unsigned char		*data;				// pointer to element data		
	void				*header;			// pointer to element header
	int					header_type;		// header type
	int					data_size;			// size of data area
	int					cb_offset;			// index to circular buffer referens
	int					cb_size;			// size of data in the circular buffer
	struct _CT_Element	*next;	
};

struct _CT_Buffer {
	struct _CT_Element	*first;	
	struct _CT_Element	*last;	
};

typedef struct _CT_Buffer  CT_Buffer;
typedef struct _CT_Element CT_Element;


CT_Buffer  *buffer_create(void);									// create a new buffer

CT_Element *buffer_add(CT_Buffer *buffer, unsigned char *data, int data_size, void *header, int header_type);

int  buffer_delete(CT_Buffer *buffer, int options);					// delete the first item in the list

void buffer_destroy(CT_Buffer *buffer);								// delete all items in the list including the buffer

int  buffer_empty(CT_Buffer *buffer);								// returns true if empty

int  buffer_size(CT_Buffer *buffer);								// returns the total size in bytes

int  buffer_count(CT_Buffer *buffer);								// return the number of elements in the buffer

int  buffer_move(CT_Buffer *from, CT_Buffer *to);					// move the first element in 'from' to 'to'

int  buffer_move_all(CT_Buffer *from, CT_Buffer *to);				// move all elements in the buffer

//void buffer_debug_info(char* name, CT_Buffer *buffer);

#endif
