#ifndef __DECODER_H__
#define __DECODER_H__

#include "constants.h"
#include "buffer.h"
#include "bitstream.h"

// samples          : a pointer to the pcm samples
// num_samples      : the number of samples in the buffer (for each channel)
// num_channels     : number of channels
// bytes_per_sample : number of bytes to represent one sample for one channel
// frequency        : the sampling frequency

typedef struct {
	CT_Buffer		*input_buffer;	
	CT_Buffer		*output_buffer;	
	CT_Bitstream	*bitstream;
	CT_Bitstream	*cb_bitstream;
	unsigned char	*cbuffer;
	unsigned int	cb_pos;			// cbuffer position
	unsigned int	cb_size;		// cbuffer size
	short pcm_sample[2][2][MAX_SS][MAX_SB];
} CT_Decoder;

typedef struct {
	short num_samples;
	short num_channels;
	short bytes_per_sample;
	int   frequency;
} CT_OutputHeader;

typedef struct {
	short num_samples;
	short num_channels;
	short bytes_per_sample;
	int   frequency;
} CT_InputHeader;

// input_open   : open the input device and return device status
// input_close  : close the input device if possible
// input_read   : blocking call that returns a pointer to the data and the size
// input_status : returns the status code. 

typedef struct {
    int  (*open)(char* filename);
    int  (*close)(void);
    int  (*read)(CT_Decoder *decoder);
    int  (*status)(void);
    int  (*ioctl)(unsigned long action, unsigned long par1, void *par2);
} CT_InputModule;

// input_open   : open the input device and return device status
// input_close  : close the input device if possible
// input_read   : blocking call that returns a pointer to the data and the size
// input_status : returns the status code. 

typedef struct {
    int  (*open)(char *filename);
    int  (*close)(void);
    int  (*write)(CT_Decoder *decoder);
    int  (*status)(void);
    int  (*ioctl)(unsigned long action, unsigned long par1, void *par2);
} CT_OutputModule;

//typedef int (*input_register)(CT_InputModule *input_module) CT_InputRegister; 
//typedef int (*output_register)(CT_OutputModule *output_module) CT_OutputRegister; 

// start the decoding process.
// a call to decoder_run() will 
// be blocked until end_of_stream.

void decoder_init(int (*input_register)(CT_InputModule *input_module), int (*output_register)(CT_OutputModule *output_module));

void decoder_run(char* filename);

#endif
