#include <stdio.h>	
#include "../main.h"

#include "../memhandler.h"
#include "../frames.h"
#include "fileinput.h"

#define BLOCK_SIZE	1024

FILE *disk_file_in;
static file_input_size = 0;

int file_input_open(char* filename) 
{
	char new_filename[200];
	
	sprintf(new_filename, "input_samples/%s", filename);
	
	if ((disk_file_in = fopen(new_filename, "rb")) == 0) {
		printf("Cannot open file: %s\n", new_filename);
		return CT_ERROR;
	} else {
		printf("File opened: %s\n", new_filename);
		return CT_OK;
	}
}

int file_input_close(void)
{
	fclose(disk_file_in);
	
	return CT_OK;
}

int file_input_read(CT_Decoder *decoder)
{
	unsigned int			num_read, i, N, hs, start_byte, end_byte, start_bit;
	static unsigned char	buff_data[4000];
	unsigned char			*header;
	CT_Bitstream *bitstream;
	CT_MPEGInfo frame_info;
	CT_Element *elem;
	int valid_frame;
	
/* 	printf("Reading file for frame data (%d)...\n", file_input_size); */
	
	if (file_input_size < 2000) {
		num_read = fread(&buff_data[file_input_size], 1, 2000-file_input_size, disk_file_in);
		if (num_read == 0) {
			return CT_OK;
		}
		file_input_size += num_read;
	}

	bitstream = bitstream_create((unsigned char*)&buff_data, 2000);

	valid_frame = 0;
	if ((start_byte = find_frame_sync(bitstream)) != -1) {
		start_bit = bitstream_pos(bitstream);
//  		printf("Valid sync found. Start bit = %d\n", start_bit); 
		if (start_bit > 8000) {
// 	 		printf("Recreate new stream with 1000 bit new data\n"); 
			// recreate the stream with the new data
			file_input_size -= 1000;
			
			for (i = 0; i < 1000; i++)
				buff_data[i] = buff_data[i+1000];

			if (file_input_size < 2000) {
				num_read = fread(&buff_data[file_input_size], 1, 2000-file_input_size, disk_file_in);
				if (num_read == 0) {
					return CT_OK;
				}
				file_input_size += num_read;
			}
			free(bitstream);
			bitstream = bitstream_create((unsigned char*)&buff_data, 2000);
			
			start_bit -= 8000;
// 	 		printf("New start bit = %d\n", start_bit); 
		}
		if (frame_header(bitstream, &frame_info)) {
			valid_frame = 1;
//  			printf("Header extracted\n"); 
			if (frame_error_check(bitstream, &frame_info)) {
//  				printf("CRC check passed\n"); 
				frame_audio_data(bitstream, &frame_info);

				hs = 12 + bitstream_pos(bitstream) - start_bit; 

				N = (144 * bitrate_table[frame_info.header.bitrate_index]) / sampling_table[frame_info.header.sampling_frequency];
				if (frame_info.header.padding_bit)
					N++;
				hs = (hs) >> 3;			// in bytes
				N -= hs;				// N - header size
				start_byte = start_byte >> 3;
				end_byte = start_byte + N + hs;
// printf("hahah9: %d, %d, %d, %d\n", start_byte, N, hs, end_byte);


				header = (unsigned char*)malloc(hs);								// allocate space for header
				memcpy(header, &bitstream->data[start_byte], hs);					// copy the header

				elem = buffer_add(decoder->input_buffer, header, hs, 0, 0);			// insert in buffer
				elem->cb_offset = decoder->cb_pos;									// index where this headers main data is put in the circular buffer
				elem->cb_size = N;													// size of data in the circular buffer

/*				{
					FILE* file;
					int i;
			
					file = fopen("debug_data/debug_bit_reservoir.txt", OUTPUT_FILE_TYPE);
					for (i = 0; i < N; i++)
						fprintf(file, "%d\n", buff_data[start_byte+hs+i]);
					fprintf(file, "\n");
					fclose(file);
				}*/

				if ((decoder->cb_size - decoder->cb_pos) >= N) {
					memcpy(&decoder->cbuffer[decoder->cb_pos], &buff_data[start_byte+hs], N);
				}
				else {
					int first = decoder->cb_size - decoder->cb_pos;
					memcpy(&decoder->cbuffer[decoder->cb_pos], &buff_data[start_byte+hs], first);
					memcpy(&decoder->cbuffer[0], &buff_data[start_byte+hs+first], N-first);
				}		
				decoder->cb_pos = (decoder->cb_pos + N) % decoder->cb_size;
						
				for (i = 0; i < 2000-end_byte; i++)
					buff_data[i] = buff_data[i+end_byte];

				file_input_size -= end_byte;
			}
/*		} else {
			hs = 12 + bitstream_pos(bitstream) - start_bit; 
			hs = hs <<3;
// 
			printf("haha: invalid header: %d, %d, %d\n", file_input_size, hs, start_bit);

			// skip the 12 sync-bit
			file_input_size -= hs;			
			
			for (i = 0; i < 2000-hs; i++) {
				buff_data[i] = buff_data[i+hs];
			}
*/		}
	}

	free(bitstream);

	if (valid_frame == 1)
		return CT_OK;
	else
		return 2;
}

int file_input_status(void)
{
	return CT_OK;
}

int file_input_ioctl(unsigned long action, unsigned long par1, void *par2)
{
	return CT_OK;
}

int file_input_register(CT_InputModule *input_module) 
{
	input_module->open   = file_input_open;
	input_module->close  = file_input_close;
	input_module->read   = file_input_read;
	input_module->status = file_input_status;
	input_module->ioctl  = file_input_ioctl;

	return CT_OK;
}
