#include <stdio.h>	
#include <string.h>
#include "../main.h"

#include "fileoutput.h"

void out_fifo(short[][][],int,int,FILE *,unsigned long *);

FILE *disk_file_out;
unsigned long sample_frames = 0;

int file_output_open(char* filename)
{
char new_filename[200], temp[200];
char *p;

p = strchr(filename, '.');
strncpy(temp, filename, p-filename);
temp[p-filename] = 0;

sprintf(new_filename, "output_samples/%s_%d.wav", temp, DATA_WIDTH);

#if DATA_WIDTH == 32
	if ((disk_file_out = fopen(new_filename, "wb")) == 0)	{
#else
	if ((disk_file_out = fopen(new_filename, "wb")) == 0)	{
#endif
		printf("Cannot open output file: %s\n", new_filename);
		return CT_ERROR;
	} else {
		printf("Output file opened: %s\n", new_filename);	
		return CT_OK;
	}
}

int file_output_close(void)
{
  wav_data_t        pcm_wav_data;
	
	#if WAVE_OUT == 1
  pcm_wav_data.numChannels       = 2;
//   pcm_wav_data.numSampleFrames   = ftell(disk_file_out)/4;
  pcm_wav_data.numSampleFrames   = sample_frames;
  pcm_wav_data.sampleSize        = 16;
  pcm_wav_data.sampleRate        = 44100;
//   pcm_wav_data.sampleRate        = sampling_table[
  
  if (wav_write_headers(disk_file_out, &pcm_wav_data) == -1) {
    fprintf(stdout,"Could not write wav headers output wave\n");
    exit(2);
  }
 	#endif
  
	fclose(disk_file_out);
	
	return CT_OK;
}

int file_output_write(CT_Decoder *decoder)
{
	unsigned char	*data;
	int				size;

 #if WAVE_OUT == 0
	while (!buffer_empty(decoder->output_buffer)) {
		data = decoder->output_buffer->first->data;
		size = decoder->output_buffer->first->data_size;
		fwrite(data, 1, size, disk_file_out);
		buffer_delete(decoder->output_buffer, CT_DELETE_ALL);
	}
	#else
		int ch, gr, i, j;
		
		if (last_frame == 1) {
			out_fifo(decoder->pcm_sample[0], 3, last_frame, disk_file_out, &sample_frames);
			out_fifo(decoder->pcm_sample[1], 3, last_frame, disk_file_out, &sample_frames);
		}	else {
			out_fifo(decoder->pcm_sample[0], 18, last_frame, disk_file_out, &sample_frames);
			out_fifo(decoder->pcm_sample[1], 18, last_frame, disk_file_out, &sample_frames);
		}
		while (!buffer_empty(decoder->output_buffer)) {
			buffer_delete(decoder->output_buffer, CT_DELETE_ALL);		
		}
	#endif
	
	return CT_OK;
}

int file_output_status(void)
{
	return CT_OK;
}

int file_output_ioctl(unsigned long action, unsigned long par1, void *par2)
{
	return CT_OK;
}

int file_output_register(CT_OutputModule *output_module) 
{
	output_module->open   = file_output_open;
	output_module->close  = file_output_close;
	output_module->write  = file_output_write;
	output_module->status = file_output_status;
	output_module->ioctl  = file_output_ioctl;

	return CT_OK;
}

unsigned int little_endian(unsigned int num){
  unsigned int val = 0;
  int ctr;

   for (ctr = 0; ctr < 4; ctr++)
    val |= ((num >> (ctr * 8)) & 0xff) << ((3-ctr) * 8) ;
   return val;
}


int wav_seek_to_sound_data(buffer)
     FILE *buffer;
{
  
  if (fseek(buffer,sizeof(riff_chunk_t) + sizeof(format_chunk_t) + sizeof(data_chunk_t) , SEEK_SET) != 0)
    return -1;

  //printf("seek to %d\n",ftell(buffer));
  return 0;
}


int wav_write_headers(buffer, wav_ptr)
     FILE *buffer;
     wav_data_t *wav_ptr;
{

  riff_chunk_t rct;
  format_chunk_t fct;
  data_chunk_t dct;
  long data_length;
  int bytes_per_sam;
 
  *(unsigned long *)rct.type = R_TYPE;
  *(unsigned long *)fct.type = F_TYPE;
  *(unsigned long *)dct.type = D_TYPE;
  *(unsigned long *)rct.file_type = WAVE;

  fct.length = little_endian(0x10);
  fct.mode = 1 << 24;
  fct.mode |= wav_ptr->numChannels << 8;
 
  bytes_per_sam = (wav_ptr->sampleSize + 7)/8 ;
  fct.rate = little_endian(wav_ptr->sampleRate);
 
  fct.bps = little_endian(bytes_per_sam * (wav_ptr->sampleRate) * wav_ptr->numChannels);
  fct.bytes_per_sam = ((bytes_per_sam * wav_ptr->numChannels)) << 8;
  fct.bits_per_sam = (bytes_per_sam * 8) << 8;
  
  data_length = bytes_per_sam * wav_ptr->numChannels * wav_ptr->numSampleFrames;
  dct.length = little_endian(data_length);
  rct.length = little_endian(data_length + sizeof(riff_chunk_t) + sizeof(format_chunk_t));
  
 
  if (fseek(buffer,0,SEEK_SET) != 0)
    return -1;

 if (fwrite(&rct,sizeof(riff_chunk_t), 1, buffer) != 1)
    return -1;

 if (fwrite(&fct,sizeof(format_chunk_t), 1, buffer) != 1)
    return -1;
 
 if (fwrite(&dct,sizeof(data_chunk_t), 1, buffer) != 1)
    return -1;

  printf("wav header size: %d, data length: %d\n",sizeof(riff_chunk_t)+sizeof(format_chunk_t)+sizeof(data_chunk_t),data_length);
  return 0;
}

void out_fifo(pcm_sample, num, done, outFile, psampFrames)
     short  pcm_sample[2][MAX_SS][MAX_SB];
     int num;
//      frame_params *fr_ps;
     int done;
     FILE *outFile;
     unsigned long *psampFrames;
{
  int i,j,l;
//   int stereo = fr_ps->stereo;
//   int sblimit = fr_ps->sblimit;
  int stereo = 2;
  short data;
  static short int outsamp[1600];
  static long k = 0;

  //printf("k = %d\n",k);
  if (!done)
    for (i=0;i<num;i++) 
    	for (j=0;j<MAX_SB;j++) {
      	(*psampFrames)++;
	      for (l=0;l<stereo;l++) {
					if (!(k%1600) && k) {
					  //printf("writing data at %d, k=%d\n",ftell(outFile),k);
					  fwrite(outsamp,2,1600,outFile);
					  k = 0;
					}
					data =  pcm_sample[l][i][j];
// 					printf("i=%d, j=%d, #=%d, l=%d, %04X\n", i, j, i+j*MAX_SS, l, data & 0x0000FFFF);					
					outsamp[k++] = ((data & 0xff) << 8) | ((data >> 8) & 0xff);
      	}
	    }
  else {
    fwrite(outsamp,2,(int)k,outFile);
    k = 0;
  }
// printf("post out pos: %d, %d\n",ftell(outFile), *psampFrames);
// getchar();
}
