// -------------------------------------------------------
//
// Author: Thomas Lenart    
// Email:  thomas.lenart@es.lth.se
// Date:   March 2003
//
// -------------------------------------------------------
#include "main.h"

#ifdef DEBUG_REQUANTIZE
#include <stdio.h>
#endif

#include "requantizer.h"

#if UNSIGNED_REQ == 1
  #define fix_mul2(x, y) (unsigned)(((unsigned short)(x))*((unsigned short)(y)))
#endif

//#define debug1 1
//#define debug1_frame 27

// #if DATA_WIDTH == 32
//   #define samplesoutput 0
// #endif

// #ifndef DEBUG_REQUANTIZE
//   #define samplesoutput 1
// #endif

struct _requantize_table {
  #if UNSIGNED_REQ == 1
  unsigned short value;
  #else
	fixed_point		value;			// the value with largest possible dynamic range
	#endif
	unsigned char	exponent;		// the exponent corresponding to the value
} requantize_table[1024] =	{
						#if NEW_CODE == 0
							#if DATA_WIDTH == 32
								#include "table_requantize"
							#else
								#include "table_requantize_16"
							#endif
						#else
							#if DATA_WIDTH == 32
								#include "table_requantizer_32_new"
							#else
                #if UNSIGNED_REQ == 1
                  #include "table_requantizer_16_unsigned"
                #else
									#include "table_requantizer_16_new"
								#endif
							#endif
						#endif
							};

static const fixed_point gain_correction[9] =	{ 
												#if DATA_WIDTH == 32
													#include "table_gain_correction"	// correct for 2^x (x = -5/4, -4/4, -3/4, -2/4, -1/4, 0, 1/4, 1/2, 3/4)
												#else
                          #if UNSIGNED_REQ == 1
													  #include "table_gain_correction_16_unsigned"	// correct for 2^x (x = -5/4, -4/4, -3/4, -2/4, -1/4, 0, 1/4, 1/2, 3/4)
                          #else
                            #include "table_gain_correction_16" // correct for 2^x (x = -5/4, -4/4, -3/4, -2/4, -1/4, 0, 1/4, 1/2, 3/4)
                          #endif
												#endif
												}; 

void frame_requantize(CT_MPEGInfo *frame_info, CT_Channel *channel_info, int gr, int ch, int is[MAX_SB*MAX_SS], fixed_point xr[MAX_SB*MAX_SS])
{
	unsigned int index, sfb, freq, next_sfb_boundary, sfb_begin, sfb_width, window;
	int is_val, exp, gain_exp, gain_index;
	fixed_point correction_factor;
//  #if UNSIGNED_REQ == 1
    unsigned mult_out;
    unsigned short mult_in1;
    unsigned short mult_in2;
    int tempexp2;
//  #endif
  #if LSB3_ADD == 1
    int is_val_LSB3;
  #endif
	int i;
  int pause;
  pause = 0;
  
  #if samplesoutput == 1
    FILE *dfile, *rfile;
    if ((dfile = fopen("debug_data/decoded_samples.txt", OUTPUT_FILE_TYPE)) == NULL) {
    	printf("Error opening file for storing requantized samples for debug\n");
    	exit(1);
    }
    if ((rfile = fopen("debug_data/requantized_samples.txt", OUTPUT_FILE_TYPE)) == NULL) {
    	printf("Error opening file for storing output requantized samples\n");
    	exit(1);
    }
  #endif
  
	#ifdef DEBUG_REQUANTIZE
		FILE* file;
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_requantize_input_32.txt", OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_requantize_input_16.txt", OUTPUT_FILE_TYPE);
		#endif
		fprintf(file, "// This is test case %d\n", test_case);
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, ch_count);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);			
		fprintf(file, "// block_type: %d\n", channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", channel_info->global_gain);

		for (i = 0; i < MAX_SB*MAX_SS; i++) {
			fprintf(file, "%d: %d (%0*X)\n", i, is[i], DATA_WIDTH/4, is[i] & MASK);
		}
		fclose(file);
		}
	#endif
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_requantize_intermediate_32.txt", OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_requantize_intermediate_16.txt", OUTPUT_FILE_TYPE);
		#endif
		fprintf(file, "// This is test case %d\n", test_case);
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, ch_count);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);			
		fprintf(file, "// block_type: %d\n", channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", channel_info->global_gain);
		
		fprintf(file, "// all scalefac_l\n// ");
		for (sfb=0;sfb<22;sfb++)
		  fprintf(file, "%d ", sfb, frame_info->scalefac_l[ch][sfb]);
		fprintf(file, "\n");
	}
	#endif

	freq = frame_info->header.sampling_frequency;
	sfb = 0;

	if (channel_info->window_switching_flag && (channel_info->block_type == 2) )
		if (channel_info->mixed_block_flag) 
			next_sfb_boundary = sfBandIndex_l[freq][1];  /* LONG blocks: 0,1,3 */
		else {
			next_sfb_boundary = sfBandIndex_s[freq][1]*3; /* pure SHORT block */
			sfb_width = sfBandIndex_s[freq][1];
			sfb_begin = 0;
		}  
	else 
		next_sfb_boundary = sfBandIndex_l[freq][1];  /* LONG blocks: 0,1,3 */

	channel_info->cbmax = 0;
	for (index = 0; index < MAX_SB*MAX_SS; index++) {
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "%d: ", index);
		}
	#endif
		if (index == next_sfb_boundary)  { /* Adjust critical band boundary */
			if (channel_info->window_switching_flag && (channel_info->block_type == 2)) {
				if (channel_info->mixed_block_flag)  {
					if (index == sfBandIndex_l[freq][8])  {
						next_sfb_boundary = sfBandIndex_s[freq][4]*3; 
						sfb = 3;
						sfb_width = sfBandIndex_s[freq][sfb+1] - sfBandIndex_s[freq][sfb];
						sfb_begin = sfBandIndex_s[freq][sfb]*3;      
					}
					else if (index < sfBandIndex_l[freq][8]) 
						next_sfb_boundary = sfBandIndex_l[freq][(++sfb)+1];
					else {
						next_sfb_boundary = sfBandIndex_s[freq][(++sfb)+1]*3;
						sfb_width = sfBandIndex_s[freq][sfb+1] - sfBandIndex_s[freq][sfb];
						sfb_begin = sfBandIndex_s[freq][sfb]*3;      
					}   
				}
				else  {
					next_sfb_boundary = sfBandIndex_s[freq][(++sfb)+1]*3;
					sfb_width = sfBandIndex_s[freq][sfb+1] - sfBandIndex_s[freq][sfb];
					sfb_begin = sfBandIndex_s[freq][sfb]*3;      
				}	 
			}
			else /* long blocks */
				next_sfb_boundary = sfBandIndex_l[freq][(++sfb)+1];
		}

		is_val = is[index];											// get the is value
		
		#ifdef DEBUG_REQUANTIZE
			if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
				fprintf(file, "is=%d: ", is[index]);
        #if samplesoutput == 1
        if (frame_count >= samplesframe)
          fprintf(dfile,"%d\n",is[index]);
        #endif
			}
		#endif

		if (is_val) {
			if (sfb > channel_info->cbmax) {
				channel_info->cbmax = sfb;
			}
			if (is_val < 0)												// calculate absolute
				is_val = 0 - is_val;

			if (is_val > 1023) {										// if in range 1024-8192
        #if LSB3_ADD == 1
          is_val_LSB3 = is_val % 8;
        #endif
				is_val >>= 3;											// change it to range 0-1023
				xr[index] = requantize_table[is_val].value;				// read out the value from the table
				exp = requantize_table[is_val].exponent + 4;			// read exponent and add 4 (= multiply by 16)		
			}
			else {
        #if LSB3_ADD == 1
          is_val_LSB3 = 0;
        #endif
				xr[index] = requantize_table[is_val].value;				// read out the value from the table
				exp = requantize_table[is_val].exponent;				// read exponent
			}
		
      #if SEPARATE_SIGN == 0
      if (is[index] < 0)                      // if original value is negative
        xr[index] = 0 - xr[index];                // make result negative
      #endif

			if (channel_info->window_switching_flag && (((channel_info->block_type == 2) && (channel_info->mixed_block_flag == 0)) || ((channel_info->block_type == 2) && channel_info->mixed_block_flag && (index >= 36)) )) {
				window = (index - sfb_begin) / sfb_width;

// 			if ((frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_MS_STEREO)) {
// 			// should check for round of zero of right channel
// 				gain_exp = channel_info->global_gain - 210 - (channel_info->subblock_gain[window] << 3) - 2;
// 			} else {
				gain_exp = channel_info->global_gain - 210 - (channel_info->subblock_gain[window] << 3);
// 			}
				gain_index = gain_exp & 0x03;
				exp += gain_exp >> 2;
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "exp_before_scalefac=%d window_num=%d ", exp, window);
		}
	#endif
				gain_exp = frame_info->scalefac_s[ch][sfb][window];
				if (!channel_info->scalefac_scale) {
					gain_index -= (gain_exp & 1) << 1;
					gain_exp >>= 1;
				}

				correction_factor = gain_correction[gain_index + 5];

				exp -= gain_exp;
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "correction_factor=%d xr=%d ", correction_factor, xr[index]);
		}
	#endif
				
//				xr[index] = fix_mul(correction_factor, xr[index]);			// !!! no need for multiplication with 1. Check this case?
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "exp_out=%d gain_index=%d mult_out=%d ", exp, gain_index, xr[index]);
      #if LSB3_ADD == 1
        fprintf(file, "is_val_LSB3=%d is_val_exp=%d ", is_val_LSB3, (14 - requantize_table[is_val].exponent));
      #endif
		}
	#endif

          #if ADJUST_EXP == 1
            exp += 3;
          #endif

        #if UNSIGNED_REQ == 1
          mult_in2 = requantize_table[is_val].value;
          #if LSB3_ADD == 1
            mult_in2 += is_val_LSB3 << (14 - requantize_table[is_val].exponent);
            fprintf(file,"new m2=%x ",mult_in2);
            //if (mult_in2 > 0xFFFF)
              //mult_in2 = 0xFFFF;
          #endif
          mult_in1 = correction_factor;
          mult_out = fix_mul2(mult_in1,mult_in2);
        
          xr[index] = mult_out >> 17;
          exp += 2;
          if (exp > 0)
            xr[index] <<= exp;
          else
            xr[index] >>= -exp;
        /*
          tempexp2 = 14 - exp;
          if (abs(tempexp2) >= 32)
            mult_in2 = 0;
          else if (tempexp2 < 0)
            mult_in2 = (mult_out << -tempexp2) & MASK;
          else
            mult_in2 = (mult_out >> tempexp2) & MASK;
          xr[index] = mult_in2 >> 1;
        */
        #else
          #if LSB3_ADD == 1
            xr[index] += is_val_LSB3 << (14 - requantize_table[is_val].exponent);
						#ifdef DEBUG_REQUANTIZE
          	  fprintf(file,"new m2=%x ",xr[index]);
          	#endif
          #endif
          xr[index] = fix_mul(correction_factor, xr[index]);     // !!! no need for multiplication with 1. Check this case?
          if (exp > 0)
            xr[index] <<= exp;
          else
            xr[index] >>= -exp;
          if (abs(exp) > 31) xr[index] = 0;
        #endif
        
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "exp_out=%d gain_index=%d mult_out=%x ", exp, gain_index, xr[index]);
      //fprintf(file, " -%x- ", xr[index] & MASK);
		}

	#endif
  
//       #endif
			}
			else {   /* LONG block types 0,1,3 & 1st 2 subbands of switched blocks */
				
// 				if ((frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_MS_STEREO)) {
// 			// should check for round of zero of right channel
// 					gain_exp = channel_info->global_gain - 210-2;
// 				} else {
					gain_exp = channel_info->global_gain - 210;
// 				}
			
				gain_index = gain_exp & 0x03;
				exp += gain_exp >> 2;
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "exp_before_scalefac=%d window_num=%d ", exp, window);
		}
	#endif
				
	#ifdef DEBUG_REQUANTIZE
		fprintf(file, "ch=%d sfb=%d scalefac_l=%d ", ch, sfb, frame_info->scalefac_l[ch][sfb]);
	#endif
	
				gain_exp = frame_info->scalefac_l[ch][sfb];
				if (channel_info->preflag)
					gain_exp += pretab[sfb];

				if (!channel_info->scalefac_scale) {
					gain_index -= (gain_exp & 1) << 1;
					gain_exp >>= 1;
				}

				correction_factor = gain_correction[gain_index + 5];

				exp -= gain_exp;
				
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "correction_factor=%d xr=%d ", correction_factor, xr[index]);
		}
	#endif
				
//				xr[index] = fix_mul(correction_factor, xr[index]);			// !!! no need for multiplication with 1. Check this case?
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "exp_out=%d gain_index=%d mult_out=%d ", exp, gain_index, xr[index]);
      #if LSB3_ADD == 1
        fprintf(file, "is_val_LSB3=%d is_val_exp=%d ", is_val_LSB3, (14 - requantize_table[is_val].exponent));
      #endif
		}
	#endif

				#if ADJUST_EXP == 1
            exp += 3;
          #endif
          
        #if UNSIGNED_REQ == 1
          mult_in2 = requantize_table[is_val].value;
          #if LSB3_ADD == 1
            mult_in2 += is_val_LSB3 << (14 - requantize_table[is_val].exponent);
            fprintf(file,"new m2=%x ",mult_in2);
            //if (mult_in2 > 0xFFFF)
            //  mult_in2 = 0xFFFF;
          #endif
          mult_in1 = correction_factor;
          mult_out = fix_mul2(mult_in1,mult_in2);
        
          xr[index] = mult_out >> 17;
          exp += 2;
          if (exp > 0)
            xr[index] <<= exp;
          else
            xr[index] >>= -exp;
        
        /*
          tempexp2 = 14 - exp;
          if (abs(tempexp2) >= 32)
            mult_in2 = 0;
          else if (tempexp2 < 0)
            mult_in2 = (mult_out << -tempexp2) & MASK;
          else
            mult_in2 = (mult_out >> tempexp2) & MASK;
          xr[index] = mult_in2 >> 1;
        */
        #else
          #if LSB3_ADD == 1
            xr[index] += is_val_LSB3 << (14 - requantize_table[is_val].exponent);
            #ifdef DEBUG_REQUANTIZE
							fprintf(file,"new m2=%x ",xr[index]);
						#endif
          #endif
          xr[index] = fix_mul(correction_factor, xr[index]);     // !!! no need for multiplication with 1. Check this case?
          if (exp > 0)
            xr[index] <<= exp;
          else
            xr[index] >>= -exp;
          if (abs(exp) > 31) xr[index] = 0;
        #endif
        
  #ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "exp_out=%d gain_index=%d mult_out=%x ", exp, gain_index, xr[index]);
		}
	#endif
          
//       #endif
			}
      
      #if SEPARATE_SIGN == 1
      if (is[index] < 0)                      // if original value is negative
        xr[index] = 0 - xr[index];                // make result negative
      #endif
		}
		else {
			xr[index] = 0;				
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "input=0 ");
		}
	#endif
		}
			

// 		#ifdef DEBUG_REQUANTIZE
//		fprintf(file, "xr[%d][%d] %.10f\n", index/18, index % 18, fixed_to_double(xr[index]));
// 			fprintf(file, "%d\n", xr[index]);
// 		#endif
	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fprintf(file, "dataout=%d (%0*X)\n", xr[index], DATA_WIDTH/4, xr[index] & MASK);
        #if samplesoutput == 1
        if (frame_count >= samplesframe)
          fprintf(rfile,"%d\n",xr[index]);
        #endif
		}
	#endif

	}

	#ifdef DEBUG_REQUANTIZE
		// for intermediate values
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fclose(file);
		}
	#endif

	#ifdef DEBUG_REQUANTIZE
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
		#if DATA_WIDTH == 32
	file = fopen("debug_data/debug_requantize_output_32.txt", OUTPUT_FILE_TYPE);
	#else
	file = fopen("debug_data/debug_requantize_output_16.txt", OUTPUT_FILE_TYPE);
	#endif
		fprintf(file, "// This is test case %d\n", test_case);
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, ch_count);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);			
		fprintf(file, "// block_type: %d\n", channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", channel_info->global_gain);
		for (i = 0; i < MAX_SB*MAX_SS; i++) {
			fprintf(file, "%d: %d (%0*X)\n", i, xr[i], DATA_WIDTH/4, xr[i] & MASK);
		}
		fclose(file);
	}
	#endif
  #if samplesoutput == 1
    fclose(rfile);
    fclose(dfile);
  #endif
}
