// -------------------------------------------------------
//
// Author: Thomas Lenart    
// Email:  thomas.lenart@es.lth.se
// Date:   March 2003
//
// -------------------------------------------------------

#include "main.h"

#ifdef DEBUG_STEREO
#include <stdio.h>
#endif

#include <string.h>
#include <math.h>
#include "stereo.h"

#ifdef DEBUG_STEREO
  #define samplesoutput 1
#endif

/* intensity stereo */
/* if ms mode quant pre-scales all values by 1.0/sqrt(2.0) ms_mode in table
   compensates   */
/* [ms_mode 0/1][sf][left/right]  */
static fixed_point lr[2][8][2];

static fixed_point ms_constant;

int f_get_cb(int i, int block_type, CT_MPEGInfo *frame_info);
int f_get_cb_is_start(int i, int block_type, CT_MPEGInfo *frame_info);
int f_get_window(int i, int cb, CT_MPEGInfo *frame_info);

void stereo_init() {
	int i;
  double s, c;
  double pi;
  double t;
  float f_constant;
  
  // Get fixed point of constant 2^-1/2	
	f_constant = 1.0/sqrt(2.0);
	ms_constant = f_adjusted_translate_number(f_constant);
//   printf("haha0: ms_constant=%0*X\n", DATA_WIDTH/4, ms_constant & MASK);

//    pi = 4.0 * atan(1.0);
	 pi = acos(-1);
   t = pi / 12.0;
   for (i = 0; i < 7; i++) {
      s = sin(i * t);
      c = cos(i * t);
    	f_constant = (float) (s / (s + c));
      lr[0][i][0] = f_adjusted_translate_number(f_constant);
//       printf("haha0: %d: %10d, %0*X, %f\n", i, lr[0][i][0], DATA_WIDTH/4, lr[0][i][0] & MASK, f_constant);
    	f_constant = (float) (c / (s + c));      
      lr[0][i][1] = f_adjusted_translate_number(f_constant);
//       printf("haha1: %d: %10d, %08X, %f\n", i, lr[0][i][1], lr[0][i][1], f_constant);
      
      /* ms_mode = 1 */
     	f_constant = (float) (sqrt(2.0) * (s / (s + c)));
      lr[1][i][0] = f_adjusted_translate_number(f_constant);
//       printf("haha2: %d: %10d, %0*X, %f\n", i, lr[1][i][0], DATA_WIDTH/4, lr[1][i][0] & MASK, f_constant);
    	f_constant = (float) (sqrt(2.0) * (c / (s + c)));      
      lr[1][i][1] = f_adjusted_translate_number(f_constant);
//       printf("haha3: %d: %10d, %0*X, %f\n", i, lr[1][i][1], DATA_WIDTH/4, lr[1][i][1] & MASK, f_constant);
   }
/* sf = 7 */
/* ms_mode = 0 */
   lr[0][i][0] = f_adjusted_translate_number(1.0);
   lr[0][i][1] = f_adjusted_translate_number(0.0);
/* ms_mode = 1 */
   lr[1][i][0] = f_adjusted_translate_number(1.0);
   lr[1][i][1] = f_adjusted_translate_number(1.0);
}

void stereo_processing(CT_MPEGInfo *frame_info, fixed_point xr[MAX_CH][MAX_SB*MAX_SS], fixed_point lr[MAX_CH][MAX_SB*MAX_SS])
{
	int stereo = frame_info->num_channels; //fr_ps->stereo;
	int ms_stereo = (frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_MS_STEREO);
	int i_stereo = (frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_INTENSITY_STEREO);
	int i;
	#ifdef DEBUG_STEREO
	FILE* file;
		#if DATA_WIDTH == 32
	file = fopen("debug_data/debug_stereo_32.txt", OUTPUT_FILE_TYPE);
	#else
	file = fopen("debug_data/debug_stereo_16.txt", OUTPUT_FILE_TYPE);
	#endif
	#endif

	if (stereo==2) 
/*		for(i=0;i<MAX_SB*MAX_SS;i++){
			lr[0][i] = xr[0][i];
			lr[1][i] = xr[1][i];
		}*/
		memcpy(lr, xr, (MAX_SB*MAX_SS) << 3);
	else {
/*		for(i=0;i<MAX_SB*MAX_SS;i++){
			lr[0][i] = xr[0][i];
			lr[1][i] = 0;
		}*/
		memcpy(lr[0], xr[0], (MAX_SB*MAX_SS) << 2);
		memcpy(lr[1], xr[0], (MAX_SB*MAX_SS) << 2);
	}

	#ifdef DEBUG_STEREO
	for (i = 0; i < 576; ++i) 
		fprintf(file, "lr[0][%d][%d] %.10f\n", i / 18, i % 18, fixed_to_double(lr[0][i]));
	for (i = 0; i < 576; ++i) 
		fprintf(file, "lr[1][%d][%d] %.10f\n", i / 18, i % 18, fixed_to_double(lr[1][i]));
	fclose(file);
	#endif
}

void stereo_processing_new(CT_MPEGInfo *frame_info, int gr, fixed_point xr[MAX_CH][MAX_SB*MAX_SS])
{
	int ms_stereo = (frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_MS_STEREO);
	int i_stereo = (frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_INTENSITY_STEREO);
	int i;
	int l_nsample, r_nsample;
	fixed_point left[MAX_SB*MAX_SS];
	fixed_point right[MAX_SB*MAX_SS];
  CT_Channel	*left_channel_info, *right_channel_info;
	int ms_limit;
	// for is_stereo
	int j, n, cb, w;
	fixed_point fl, fr;
	int m;
	int isf;
	fixed_point fls[3], frs[3];
	int cb0;

    #if samplesoutput == 1
    FILE *sfile;
    #endif

	#ifdef DEBUG_STEREO
		FILE* file;
		if (frame_count > 245)
			match_criteria = 1;

		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {

		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_stereo_input_32.txt", OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_stereo_input_16.txt", OUTPUT_FILE_TYPE);
		#endif
		fprintf(file, "// This is test case %d\n", test_case);
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, 0);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);
	  left_channel_info = &frame_info->gr[gr].ch[0];	    
		fprintf(file, "// Ch 0\n");
		fprintf(file, "// block_type: %d\n", left_channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", left_channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", left_channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", left_channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", left_channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", left_channel_info->global_gain);
		fprintf(file, "// count1: %d\n", left_channel_info->count1);
		fprintf(file, "// xr:\n");
//		for (i = 0; i < MAX_SB*MAX_SS; ++i) 
//			fprintf(file, "%d: %d (%0*X)\n", i, xr[0][i], DATA_WIDTH/4, xr[0][i] & MASK);
		fprintf(file, "// scale factor:\n");
		if (left_channel_info->block_type != 2) {
		#if NEW_STEREO == 0
			for (i = 0; i < MAX_SS; ++i) 
		#else
      for (i = 0; i < 21; ++i) 
    #endif
				fprintf(file, "cb=%d sf=%d\n", i, frame_info->scalefac_l[0][i]);
		} else {
			#if NEW_STEREO == 0
				for (i = 0; i < MAX_SS; ++i) 
			#else
	      for (i = 0; i < 12; ++i) 
	    #endif
				for (j = 0; j < 3; j++)
					fprintf(file, "cb=%d w=%d sf=%d\n", i, j, frame_info->scalefac_s[0][i][j]);	
		}

		fprintf(file, "\n");
		fprintf(file, "// Ch 1\n");
		fprintf(file, "// This is test case %d\n", test_case);			
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, 1);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);		
		right_channel_info = &frame_info->gr[gr].ch[1];	    
		fprintf(file, "// block_type: %d\n", right_channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", right_channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", right_channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", right_channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", right_channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", right_channel_info->global_gain);
		fprintf(file, "// count1: %d\n", right_channel_info->count1);
		fprintf(file, "// xr:\n");
//		for (i = 0; i < MAX_SB*MAX_SS; ++i) 
//			fprintf(file, "%d: %d (%0*X)\n", i, xr[1][i], DATA_WIDTH/4, xr[1][i] & MASK);
		fprintf(file, "// scale factor:\n");
		if (right_channel_info->block_type != 2) {
			#if NEW_STEREO == 0
      	for (i = 0; i < MAX_SS; ++i) 
      #else
	      for (i = 0; i < 21; ++i) 
	    #endif
				fprintf(file, "cb=%d sf=%d\n", i, frame_info->scalefac_l[1][i]);
		} else {
			#if NEW_STEREO == 0
      	for (i = 0; i < MAX_SS; ++i) 
      #else
	      for (i = 0; i < 12; ++i) 
	    #endif
				for (j = 0; j < 3; j++)
					fprintf(file, "cb=%d w=%d sf=%d\n", i, j, frame_info->scalefac_s[1][i][j]);	
		}
	
		fclose(file);
		}
	#endif
	#ifdef DEBUG_STEREO
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_stereo_intermediate_32.txt", OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_stereo_intermediate_16.txt", OUTPUT_FILE_TYPE);
		#endif
		fprintf(file, "// This is test case %d\n", test_case);
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, 0);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);
	  left_channel_info = &frame_info->gr[gr].ch[0];	    
		fprintf(file, "// Ch 0\n");
		fprintf(file, "// block_type: %d\n", left_channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", left_channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", left_channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", left_channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", left_channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", left_channel_info->global_gain);
		fprintf(file, "// count1: %d\n", left_channel_info->count1);
		fprintf(file, "\n");		
		fprintf(file, "// Ch 1\n");
		fprintf(file, "// This is test case %d\n", test_case);			
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, 1);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);		
		right_channel_info = &frame_info->gr[gr].ch[1];	    
		fprintf(file, "// block_type: %d\n", right_channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", right_channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", right_channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", right_channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", right_channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", right_channel_info->global_gain);
		fprintf(file, "// count1: %d\n", right_channel_info->count1);
	}
	#endif

  left_channel_info = &frame_info->gr[gr].ch[0];	    
 	l_nsample = left_channel_info->count1;

  right_channel_info = &frame_info->gr[gr].ch[1];	    
 	r_nsample = right_channel_info->count1;

	ms_limit = r_nsample;

	// stop MS until the end of the last valid sb
	if (right_channel_info->window_switching_flag && (right_channel_info->block_type == 2)) {
		#if NEW_STEREO == 0
			cb = f_get_cb(r_nsample, 2, frame_info);
		#else
			cb = f_get_cb_is_start(r_nsample, 2, frame_info);
		#endif
//  	  n = sfBandIndex_s[frame_info->header.sampling_frequency][cb+1] - sfBandIndex_s[frame_info->header.sampling_frequency][cb];
// 		ms_limit = 3*sfBandIndex_s[frame_info->header.sampling_frequency][cb] + 3*n;
		ms_limit = 3*sfBandIndex_s[frame_info->header.sampling_frequency][cb+1];
	} else {
		#if NEW_STEREO == 0
			cb = f_get_cb(r_nsample, 0, frame_info);
		#else
			cb = f_get_cb_is_start(r_nsample, 0, frame_info);
		#endif
//     n = sfBandIndex_l[frame_info->header.sampling_frequency][cb+1] - sfBandIndex_l[frame_info->header.sampling_frequency][cb];
// 		ms_limit = sfBandIndex_l[frame_info->header.sampling_frequency][cb] + n;
		ms_limit = sfBandIndex_l[frame_info->header.sampling_frequency][cb+1];
 	}

	if (ms_stereo) {
		ms_count++;
		
		for (i = 0; i < ms_limit; i++) {
			left[i] = xr[0][i] + xr[1][i];
			left[i] = fix_mul(left[i], ms_constant);
    	right[i] = xr[0][i] - xr[1][i];
			right[i] = fix_mul(right[i], ms_constant);
		#ifdef DEBUG_STEREO
			if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
				fprintf(file, "MS %d: xr[0]=%0*X, xr[1]=%0*X, ms_constant=%0*X, left=%0*X, right=%0*X\n", i, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK, DATA_WIDTH/4, ms_constant & MASK, DATA_WIDTH/4, left[i] & MASK, DATA_WIDTH/4, right[i] & MASK);
			}
		#endif
	    xr[0][i] = left[i];
  	  xr[1][i] = right[i];
  	}
	}
	
	if (i_stereo) {
		is_count++;
// 		printf("IS_mode started...\n");
		
		if (ms_stereo == 0) {
			for (i = 0; i < ms_limit; i++) {
				#ifdef DEBUG_STEREO
					if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
						fprintf(file, "BYPASS %d: xr[0]=%0*X, xr[1]=%0*X\n", i, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
					}
				#endif
			}
		}
		if (right_channel_info->window_switching_flag && (right_channel_info->block_type == 2)) {
		// short block
			for (i = ms_limit; i < 576; i++) {
				cb = f_get_cb(i, 2, frame_info);
				w = f_get_window(i, cb, frame_info);
 				isf = frame_info->scalefac_s[1][cb][w];
 				#ifdef DEBUG_STEREO
					if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
						fprintf(file, "IS %d: cb=%d, w=%d, is_pos=%d, xr[0]=%0*X, xr[1]=%0*X, ", i, cb, w, isf, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
					}
				#endif
 				switch (isf) {
 				case 0:
 					xr[1][i] = xr[0][i];
 					xr[0][i] = 0;
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "left=%0*X, right=%0*X\n", DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif
 				break;
 				case 1:
 				case 2:
					fl = lr[0][isf][0];
					fr = lr[0][isf][1];
					xr[1][i] = fix_mul(fr, xr[0][i]);
// 					xr[0][i] = fix_mul(fl, xr[0][i]);
					xr[0][i] = xr[0][i] - xr[1][i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "fl=%0*X, fr=%0*X, left=%0*X, right=%0*X\n", DATA_WIDTH/4, fl & MASK, DATA_WIDTH/4, fr & MASK, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif					
				break;
 				case 3:
 					xr[1][i] = xr[0][i] >> 1;
 					xr[0][i] = xr[1][i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "left=%0*X, right=%0*X\n", DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif
 				break;
 				case 4:
					fl = lr[0][2][0];
					fr = lr[0][2][1];
					right[i] = fix_mul(fr, xr[0][i]);
//					left[i] = fix_mul(fl, xr[0][i]);
					left[i] = xr[0][i] - right[i];

					xr[1][i] = left[i];
					xr[0][i] = right[i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "fl=%0*X, fr=%0*X, left=%0*X, right=%0*X\n", DATA_WIDTH/4, fl & MASK, DATA_WIDTH/4, fr & MASK, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif					
				break;
 				case 5:
					fl = lr[0][1][0];
					fr = lr[0][1][1];
					right[i] = fix_mul(fr, xr[0][i]);
// 					left[i] = fix_mul(fl, xr[0][i]);
					left[i] = xr[0][i] - right[i];
					xr[1][i] = left[i];
					xr[0][i] = right[i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "fl=%0*X, fr=%0*X, left=%0*X, right=%0*X\n", DATA_WIDTH/4, fl & MASK, DATA_WIDTH/4, fr & MASK, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif					
 				break;
 				case 6:
 				case 7:
 				default:
 					xr[1][i] = 0;
 					xr[0][i] = xr[0][i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "left=%0*X, right=%0*X\n", DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif
 				break;
/* 				default: 
					printf("IS %d: r_nsample=%d cb=%d, w=%d, is_pos=%d, xr[0]=%0*X, xr[1]=%0*X, ", i, r_nsample, cb, w, isf, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
 					printf("Error: IS_POS (%d) out of range...\n", isf);
// 					getchar();
*/
				}
			}
		} else {
			for (i = ms_limit; i < 576; i++) {
				cb = f_get_cb(i, 0, frame_info);
 				isf = frame_info->scalefac_l[1][cb];
 				#ifdef DEBUG_STEREO
					if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
						fprintf(file, "IS %d: cb=%d, is_pos=%d, xr[0]=%0*X, xr[1]=%0*X, ", i, cb, isf, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
					}
				#endif

 				switch (isf) {
 				case 0:
 					xr[1][i] = xr[0][i];
 					xr[0][i] = 0;
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "left=%0*X, right=%0*X\n", DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif
 				break;
 				case 1:
 				case 2:
					fl = lr[0][isf][0];
					fr = lr[0][isf][1];
					xr[1][i] = fix_mul(fr, xr[0][i]);
// 					xr[0][i] = fix_mul(fl, xr[0][i]);
					xr[0][i] = xr[0][i] - xr[1][i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "fl=%0*X, fr=%0*X, left=%0*X, right=%0*X\n", DATA_WIDTH/4, fl & MASK, DATA_WIDTH/4, fr & MASK, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif		
				break;
 				case 3:
 					xr[1][i] = xr[0][i] >> 1;
 					xr[0][i] = xr[1][i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "left=%0*X, right=%0*X\n", DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif
 				break;
 				case 4:
					fl = lr[0][2][0];
					fr = lr[0][2][1];
					right[i] = fix_mul(fr, xr[0][i]);
// 					left[i] = fix_mul(fl, xr[0][i]);
					left[i] = xr[0][i] - right[i];
					xr[1][i] = left[i];
					xr[0][i] = right[i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "fl=%0*X, fr=%0*X, left=%0*X, right=%0*X\n", DATA_WIDTH/4, fl & MASK, DATA_WIDTH/4, fr & MASK, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif					
				break;
 				case 5:
					fl = lr[0][1][0];
					fr = lr[0][1][1];
					right[i] = fix_mul(fr, xr[0][i]);
// 					left[i] = fix_mul(fl, xr[0][i]);
					left[i] = xr[0][i] - right[i];
					xr[1][i] = left[i];
					xr[0][i] = right[i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "fl=%0*X, fr=%0*X, left=%0*X, right=%0*X\n", DATA_WIDTH/4, fl & MASK, DATA_WIDTH/4, fr & MASK, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif					
 				break;
 				case 6:
 				case 7:
 				default:
 					xr[1][i] = 0;
 					xr[0][i] = xr[0][i];
	 				#ifdef DEBUG_STEREO
						if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
							fprintf(file, "left=%0*X, right=%0*X\n", DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
						}
					#endif
 				break;
/* 				default: 
					printf("IS %d: r_nsample=%d cb=%d, is_pos=%d, xr[0]=%0*X, xr[1]=%0*X\n", i, r_nsample, cb, isf, DATA_WIDTH/4, xr[0][i] & MASK, DATA_WIDTH/4, xr[1][i] & MASK);
 					printf("Error: IS_POS (%d) out of range...\n", isf);
// 					getchar();
*/
				}
			}
		}
	}
	
	#ifdef DEBUG_STEREO
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
			fclose(file);
		}
	#endif

	#ifdef DEBUG_STEREO
		if ((ch_count == TARGET_CHANNEL || STEREO == 1) && (match_criteria == 1)) {
    #if samplesoutput == 1
      sfile = fopen("debug_data/stereo_samples.txt", OUTPUT_FILE_TYPE);
    #endif
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_stereo_output_32.txt", OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_stereo_output_16.txt", OUTPUT_FILE_TYPE);
		#endif
				fprintf(file, "// This is test case %d\n", test_case);
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, 0);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);
	  left_channel_info = &frame_info->gr[gr].ch[0];	    
		fprintf(file, "// Ch 0\n");
		fprintf(file, "// block_type: %d\n", left_channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", left_channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", left_channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", left_channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", left_channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", left_channel_info->global_gain);
		fprintf(file, "// count1: %d\n", left_channel_info->count1);
		for (i = 0; i < MAX_SB*MAX_SS; ++i) 
			fprintf(file, "%d: %d (%0*X)\n", i, xr[0][i], DATA_WIDTH/4, xr[0][i] & MASK);

		fprintf(file, "\n");
		fprintf(file, "// Ch 1\n");
		fprintf(file, "// This is test case %d\n", test_case);			
		fprintf(file, "// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, 1);
		fprintf(file, "// sampling_frequency: %d\n", frame_info->header.sampling_frequency);
		fprintf(file, "// mode: %d\n", frame_info->header.mode);
		fprintf(file, "// mode ext: %d\n", frame_info->header.mode_extension);		
		right_channel_info = &frame_info->gr[gr].ch[1];	    
		fprintf(file, "// block_type: %d\n", right_channel_info->block_type);
		fprintf(file, "// window_switching_flag: %d\n", right_channel_info->window_switching_flag);
		fprintf(file, "// mixed_block_flag: %d\n", right_channel_info->mixed_block_flag);
		fprintf(file, "// scalefac_scale: %d\n", right_channel_info->scalefac_scale);
		fprintf(file, "// preflag: %d\n", right_channel_info->preflag);
		fprintf(file, "// global_gain: %d\n", right_channel_info->global_gain);
		fprintf(file, "// count1: %d\n", right_channel_info->count1);
		for (i = 0; i < MAX_SB*MAX_SS; ++i) 
			fprintf(file, "%d: %d (%0*X)\n", i, xr[1][i], DATA_WIDTH/4, xr[1][i] & MASK);
    fclose(file);
    #if samplesoutput == 1
      if (frame_count >= samplesframe)
        for (i = 0; i < MAX_SB*MAX_SS; ++i) 
          fprintf(sfile, "%d\n%d\n",xr[1][i],xr[0][i]);
      fclose(sfile);
    #endif
		}
	#endif
}

int f_get_cb(int i, int block_type, CT_MPEGInfo *frame_info) {
  // getting cb for each sample
	int cb;
	
	cb = 0;
  
  if (block_type == 2) {
    while (i >= 3*sfBandIndex_s[frame_info->header.sampling_frequency][cb]) cb++;
  } else {
    while (i >= sfBandIndex_l[frame_info->header.sampling_frequency][cb]) cb++;
  }
	
	// return the scale factor band where i resides.
	return cb-1;
}

int f_get_cb_is_start(int i, int block_type, CT_MPEGInfo *frame_info) {
  // getting cb for each sample
  int cb;
  
  cb = 0;
  
  if (block_type == 2) {
    while (i > 3*sfBandIndex_s[frame_info->header.sampling_frequency][cb+1]) cb++;
  } else {
    while (i > sfBandIndex_l[frame_info->header.sampling_frequency][cb+1]) cb++;
  }
  
  // return the scale factor band where i resides.
  return cb;
}

int f_get_window(int i, int cb, CT_MPEGInfo *frame_info) {
	int temp, window;
	
// 	cb_width = sfBandIndex_s[frame_info->header.sampling_frequency][cb+1] - sfBandIndex_s[frame_info->header.sampling_frequency][cb];
// 	temp = i - sfBandIndex_s[frame_info->header.sampling_frequency][cb];
// 	window = temp % 3;

	window = i % 3;
	
	// return the scale factor band where i resides.
	return window;
}
