/**********************************************************************
Copyright (c) 1991 MPEG/audio software simulation group, All Rights Reserved
musicout.c
**********************************************************************/
/**********************************************************************
 * MPEG/audio coding/decoding software, work in progress              *
 *   NOT for public distribution until verified and approved by the   *
 *   MPEG/audio committee.  For further information, please contact   *
 *   Chad Fogg email: <cfogg@xenon.com>                               *
 *                                                                    *
 * VERSION 4.1                                                        *
 *   changes made since last update:                                  *
 *   date   programmers                comment                        *
 * 2/25/91  Douglas Wong        start of version 1.0 records          *
 * 3/06/91  Douglas Wong        rename setup.h to dedef.h             *
 *                              removed extraneous variables          *
 *                              removed window_samples (now part of   *
 *                              filter_samples)                       *
 * 3/07/91  Davis Pan           changed output file to "codmusic"     *
 * 5/10/91  Vish (PRISM)        Ported to Macintosh and Unix.         *
 *                              Incorporated new "out_fifo()" which   *
 *                              writes out last incomplete buffer.    *
 *                              Incorporated all AIFF routines which  *
 *                              are also compatible with SUN.         *
 *                              Incorporated user interface for       *
 *                              specifying sound file names.          *
 *                              Also incorporated user interface for  *
 *                              writing AIFF compatible sound files.  *
 * 27jun91  dpwe (Aware)        Added musicout and &sample_frames as  *
 *                              args to out_fifo (were glob refs).    *
 *                              Used new 'frame_params' struct.       *
 *                              Clean,simplify, track clipped output  *
 *                              and total bits/frame received.        *
 * 7/10/91  Earle Jennings      changed to floats to FLOAT            *
 *10/ 1/91  S.I. Sudharsanan,   Ported to IBM AIX platform.           *
 *          Don H. Lee,                                               *
 *          Peter W. Farrett                                          *
 *10/ 3/91  Don H. Lee          implemented CRC-16 error protection   *
 *                              newly introduced functions are        *
 *                              buffer_CRC and recover_CRC_error      *
 *                              Additions and revisions are marked    *
 *                              with "dhl" for clarity                *
 * 2/11/92  W. Joseph Carter    Ported new code to Macintosh.  Most   *
 *                              important fixes involved changing     *
 *                              16-bit ints to long or unsigned in    *
 *                              bit alloc routines for quant of 65535 *
 *                              and passing proper function args.     *
 *                              Removed "Other Joint Stereo" option   *
 *                              and made bitrate be total channel     *
 *                              bitrate, irrespective of the mode.    *
 *                              Fixed many small bugs & reorganized.  *
 *19 aug 92 Soren H. Nielsen    Changed MS-DOS file name extensions.  *
 * 8/27/93 Seymour Shlien,      Fixes in Unix and MSDOS ports,        *
 *         Daniel Lauzon, and                                         *
 *         Bill Truerniet                                             *
 *--------------------------------------------------------------------*
 * 4/23/92  J. Pineda           Added code for layer III.  LayerIII   *
 *          Amit Gulati         decoding is currently performed in    *
 *                              two-passes for ease of sideinfo and   *
 *                              maindata buffering and decoding.      *
 *                              The second (computation) pass is      *
 *                              activated with "decode -3 <outfile>"  *
 * 10/25/92 Amit Gulati         Modified usage() for layerIII         *
 * 12/10/92 Amit Gulati         Changed processing order of re-order- *
 *                              -ing step.  Fixed adjustment of       *
 *                              main_data_end pointer to exclude      *
 *                              side information.                     *
 *  9/07/93 Toshiyuki Ishino    Integrated Layer III with Ver 3.9.    *
 *--------------------------------------------------------------------*
 * 11/20/93 Masahiro Iwadare    Integrated Layer III with Ver 4.0.    *
 *--------------------------------------------------------------------*
 *  7/14/94 Juergen Koller      Bug fixes in Layer III code           *
 *--------------------------------------------------------------------*
 *  1/15/02 Christopher Andrews Cleaned up for 545 project usage:     *
 *                              Removed all support for other layers  *
 *                              and other platforms                   *
 **********************************************************************/

#include <stdlib.h>
#include <pthread.h>
#include <errno.h>

#include "custom.h"
#include "common.h"
#include "decoder.h"

#define SWEDISH_PATH "../MP3_Lund"


/* Global variable definitions for "musicout.c" */
typedef short PCM[2][SSLIMIT][SBLIMIT];
typedef unsigned int SAM[2][3][SBLIMIT];
typedef double FRA[2][3][SBLIMIT];
typedef double VE[2][HAN_SIZE];


struct thread_args
{
  int gr;

  int stereo;
  III_scalefac_t* III_scalefac;
  III_side_info_t* III_side_info;  
  frame_params* fr_ps;
  PCM *pcm_sample;

  double lr[2][SBLIMIT][SSLIMIT];
  double ro[2][SBLIMIT][SSLIMIT];

  int clip;
};

struct thread_args* args_0;
struct thread_args* args_1;


char *programName;
int  main_data_slots();
int  side_info_slots();
void serial_decode(struct thread_args* args);
void parallel_decode(struct thread_args* args);

/* Implementations */
main(int argc, char** argv)
{   
  PCM  *pcm_sample_0; 
  PCM  *pcm_sample_1;
  SAM  *sample;
  FRA  *fraction;
  
  VE  *w;

  Bit_stream_struc  bs;
  frame_params      fr_ps;
  layer             info;
  FILE              *musicout;
  unsigned long     sample_frames;

  int               i, j, k, ch, stereo, done=FALSE, clip, sync;
  int               error_protection, crc_error_count, total_error_count;
  unsigned int      old_crc, new_crc;
  unsigned int      bit_alloc[2][SBLIMIT], scfsi[2][SBLIMIT], scale_index[2][3][SBLIMIT];
  unsigned long     bitsPerSlot, samplesPerFrame;
  unsigned long     frameBits, gotBits = 0;
 
  wav_data_t        pcm_wav_data;
  char              encoded_file_name[MAX_NAME_SIZE];
  char              decoded_file_name[MAX_NAME_SIZE];
  char*             ext;  
  char              t[50];
 
  III_scalefac_t    III_scalefac_0, III_scalefac_1;
  III_side_info_t   III_side_info_0, III_side_info_1;
  int nSlots;
  int main_data_end, flush_main;
  int  bytes_to_discard ;
  static int frame_start = 0;

 
  /* Most large variables are declared dynamically to ensure
       compatibility with smaller machines */

  pcm_sample_0 = (PCM  *) mem_alloc((long) sizeof(PCM), "PCM Samp");
  pcm_sample_1 = (PCM  *) mem_alloc((long) sizeof(PCM), "PCM Samp");
  sample = (SAM  *) mem_alloc((long) sizeof(SAM), "Sample");
  fraction = (FRA  *) mem_alloc((long) sizeof(FRA), "fraction");
  w = (VE  *) mem_alloc((long) sizeof(VE), "w");

  fr_ps.header = &info;

  for (i=0;i<HAN_SIZE;i++) 
    for (j=0;j<2;j++) 
      (*w)[j][i] = 0.0;

  programName = argv[0];

  if(argc==1) {
    fprintf(stderr,"Usage: %s <filename>\n",programName);
    exit(1);
  }

//   strcpy(encoded_file_name,argv[1]);
	sprintf(encoded_file_name, "%s/out/input_samples/%s.mp3", SWEDISH_PATH, argv[1]);
	sprintf(decoded_file_name, "%s/out/output_samples/%s.mp3", SWEDISH_PATH, argv[1]);
	
//   strcpy(decoded_file_name, encoded_file_name);
  if ((ext = strstr(decoded_file_name,MP3_OPEXT)) != NULL)
    strncpy(ext,DFLT_OPEXT,4);
  else
    strcat (decoded_file_name, DFLT_OPEXT);
  
  /* report results of command line */
  fprintf(stdout,"Input file = '%s'  output file = '%s'\n",
	  encoded_file_name, decoded_file_name);
  fprintf(stdout,"Output file written in WAV format\n");
 

 
  if ((musicout = fopen(decoded_file_name, "w+b")) == NULL) {
    printf ("Could not create \"%s\".\n", decoded_file_name);
    exit(1);
  }

  open_bit_stream_r(&bs, encoded_file_name, BUFFER_SIZE);
   
  if (wav_seek_to_sound_data(musicout) == -1) {
    fprintf(stdout,"Could not seek to PCM sound data in \"%s\".\n", decoded_file_name);
    exit(1);
  }

// 	if ((temp_fp = fopen("temp.txt", "w")) == NULL) {
//     printf ("Could not create \"temp.txt\"\n");
//     exit(1);		
// 	}
char filename[200];
sprintf(filename, "%s_left.txt", argv[1]);
printf("Opening \"%s\" to store huffman decoded samples for left channel.\n", filename);
	if ((left_data = fopen(filename, "w")) == NULL) {
    printf ("Could not create \"%s\"\n", filename);
    exit(1);		
	}
sprintf(filename, "%s_right.txt", argv[1]);
printf("Opening \"%s\" to store huffman decoded samples for right channel.\n", filename);
	if ((right_data = fopen(filename, "w")) == NULL) {
    printf ("Could not create \"%s\"\n", filename);
    exit(1);		
	}
	
  sample_frames = 0;
  frameNum = 0;
  mixed_block_count = 0;

  /*
    Main Decoding Loop
  */
  while (!end_bs(&bs)) 
  {
    sync = seek_sync(&bs, SYNC_WORD, SYNC_WORD_LNGTH);
    frameBits = sstell(&bs) - gotBits;
    if(frameNum > 0)        /* don't want to print on 1st loop; no lay */
      gotBits += frameBits;

    if (!sync) 
    {
      fprintf(stdout,"Frame cannot be located\n");
      fprintf(stdout,"Input stream may be empty\n");
      done = TRUE;
      /* finally write out the buffer */
      if (info.lay != 1) 
      {
	out_fifo(*pcm_sample_0, 3, &fr_ps, done, musicout, &sample_frames);
	out_fifo(*pcm_sample_1, 3, &fr_ps, done, musicout, &sample_frames);
      }     
      else 
      {              
	out_fifo(*pcm_sample_0, 1, &fr_ps, done, musicout, &sample_frames);
	out_fifo(*pcm_sample_1, 1, &fr_ps, done, musicout, &sample_frames);
      }
      break;
    }

    decode_info(&bs, &fr_ps);
    hdr_to_frps(&fr_ps);
    stereo = fr_ps.stereo;
    error_protection = info.error_protection;
    crc_error_count = 0;
    total_error_count = 0;
    if(frameNum == 0) /* printout layer/mode */
      WriteHdr(&fr_ps, stdout);  

    fprintf(stderr, "{%4lu}", frameNum);
    frameNum++; 
    fflush(stderr);
    if (error_protection) 
      buffer_CRC(&bs, &old_crc);

    if (info.lay != 3)
    {
      fprintf(stderr,"ERROR: Input file is not a layer III file.\n");
      exit(-1);
    }
            
    bitsPerSlot = 8;        
    samplesPerFrame = 1152;

    III_get_side_info(&bs, &III_side_info_0, &fr_ps);
    memcpy(&III_side_info_1, &III_side_info_0, sizeof(III_side_info_0));
    /*III_side_info_1 = III_side_info_0;*/

    nSlots = main_data_slots(fr_ps);
             
    for (; nSlots > 0; nSlots--)  /* read main data. */
      hputbuf((unsigned int) getbits(&bs,8), 8);
                
    //use previous bits that are left over in last frame
    main_data_end = hsstell() / 8; /*of previous frame*/
             
    if ( flush_main=(hsstell() % bitsPerSlot) ) 
    { 
      hgetbits((int)(bitsPerSlot - flush_main));
      main_data_end ++;
    }
    bytes_to_discard = frame_start - main_data_end - III_side_info_0.main_data_begin ;
                                    
    if( main_data_end > 4096 )
    {   
      frame_start -= 4096;
      rewindNbytes( 4096 );
    }

    frame_start += main_data_slots(fr_ps);
    if (bytes_to_discard < 0) {
      fprintf(stdout,"Not enough main data to decode frame %d.  Frame discarded.\n", frameNum - 1); 
//       break;
			continue;
    }
    for (; bytes_to_discard > 0; bytes_to_discard--) 
      hgetbits(8);

    clip = 0;
    /* 
       ===============================================
       >> BEGIN PARALLELIZE THIS 
       ===============================================
    */
    {
      pthread_t           tid_0;
      pthread_t           tid_1;
      
      args_0 = (struct thread_args*)malloc(sizeof(struct thread_args));
      args_1 = (struct thread_args*)malloc(sizeof(struct thread_args));     
            
      /* fill out argumnent structures */
      args_0->gr =             0;                           args_1->gr =             1;
      args_0->stereo =         stereo;                      args_1->stereo =         stereo;
      args_0->III_scalefac =   &III_scalefac_0;             args_1->III_scalefac =   &III_scalefac_1;
      args_0->III_side_info =  &III_side_info_0;            args_1->III_side_info =  &III_side_info_1;
      args_0->fr_ps =          &fr_ps;                      args_1->fr_ps =          &fr_ps;
      args_0->pcm_sample =     pcm_sample_0;                args_1->pcm_sample =     pcm_sample_1;
 

      /* Reading from the bit resevoir must be done in serial */
      serial_decode(args_0);
      serial_decode(args_1);

      /* spawn a thread for each granule */
      pthread_create( &tid_0, NULL, (void*)parallel_decode, args_0 );
      pthread_create( &tid_1, NULL, (void*)parallel_decode, args_1 );

      pthread_join(tid_0,NULL); 
      pthread_join(tid_1,NULL);

      out_fifo(*(args_0->pcm_sample), 18, &fr_ps, done, musicout, &sample_frames);
      out_fifo(*(args_1->pcm_sample), 18, &fr_ps, done, musicout, &sample_frames);
      

      clip = args_0->clip + args_1->clip;
    }
    /*
      ===============================================
      << END PARALLELIZE THIS 
      ===============================================
    */

    if(clip > 0) 
      fprintf(stdout,"%d samples clipped.\n", clip);    
  }
  
  fprintf(stdout,"\n\nwritten to %d\n",ftell(musicout));
 

  pcm_wav_data.numChannels       = stereo;
  pcm_wav_data.numSampleFrames   = sample_frames;
  printf("Sample Frames: %d\n",sample_frames);
  pcm_wav_data.sampleSize        = 16;
  pcm_wav_data.sampleRate        = s_freq[info.sampling_frequency]*1000;
  
  if (wav_write_headers(musicout, &pcm_wav_data) == -1) {
    fprintf(stdout,"Could not write wav headers to \"%s\"\n",
	    decoded_file_name);
    exit(2);
  }

  fprintf(stdout,"Avg slots/frame = %.3f; b/smp = %.2f; br = %.3f kbps\n",
	  (FLOAT) gotBits / (frameNum * bitsPerSlot),
	  (FLOAT) gotBits / (frameNum * samplesPerFrame),
	  (FLOAT) gotBits / (frameNum * samplesPerFrame) *
	  s_freq[info.sampling_frequency]);

  close_bit_stream_r(&bs);
  fclose(musicout);


  fprintf(stdout,"Decoding of \"%s\" is finished\n", encoded_file_name);
  fprintf(stdout,"The decoded PCM output file name is \"%s\"\n", decoded_file_name);
  fprintf(stdout, "Mixed block count = %d\n", mixed_block_count);
  if (mixed_block_count > 0) getchar();
  
//   fclose(temp_fp);
  fclose(left_data);
  fclose(right_data);

  exit( 0 );
}


/* ============================================================ */

/*
  SERIAL
*/
void serial_decode(struct thread_args* args)
{
  int ch;
  int stereo = args->stereo;
  double tmp, tmp2;
  double ro[2][SBLIMIT][SSLIMIT];

  for (ch=0; ch<stereo; ch++) 
  {
    long int is[SBLIMIT][SSLIMIT]; 
    int part2_start;
    part2_start = hsstell();
    III_get_scale_factors(*(args->III_scalefac),args->III_side_info, args->gr,ch, args->fr_ps);
    //tmp = ro[0][0][0];
    III_hufman_decode(is, args->III_side_info, ch, args->gr, part2_start, args->fr_ps);
    //tmp2 = ro[0][0][0];
    //if( tmp != tmp2 )
    //{
    //   printf("%lf != %lf\n",tmp,tmp2);
    //   exit(0);
    // }
    III_dequantize_sample(is, ro[ch], args->III_scalefac, &(args->III_side_info->ch[ch].gr[args->gr]), ch, args->fr_ps);
  }
  III_stereo(ro,args->lr,*(args->III_scalefac), &(args->III_side_info->ch[0].gr[args->gr]), args->fr_ps);
}
/*
  PARALLEL
*/
void parallel_decode(struct thread_args* args)
{
  int gr, ch, ss, sb, i;
  int clip = 0;
  
  int stereo =                     args->stereo;
  III_scalefac_t*  III_scalefac =  args->III_scalefac;
  III_side_info_t* III_side_info = args->III_side_info;  
  frame_params* fr_ps =            args->fr_ps;
  PCM *pcm_sample =                args->pcm_sample;
      
  gr = args->gr;
  for (ch=0; ch<stereo; ch++) 
  {
    double      re[SBLIMIT][SSLIMIT];
    double      hybridIn[SBLIMIT][SSLIMIT];/* Hybrid filter input */
    double      hybridOut[SBLIMIT][SSLIMIT];/* Hybrid filter out */
    double      polyPhaseIn[SBLIMIT];     /* PolyPhase Input. */
    mad_fixed_t polyPhaseIn_MAD[SBLIMIT];     /* PolyPhase Input. */
    
    III_reorder ((args->lr)[ch],re,&(III_side_info->ch[ch].gr[gr]), fr_ps);	
    III_antialias(re, hybridIn, &(III_side_info->ch[ch].gr[gr]), fr_ps); /* Antialias butterflies. */
    
    /* Hybrid synthesis. */
    for (sb=0; sb<SBLIMIT; sb++)  
    {
      III_hybrid(hybridIn[sb], hybridOut[sb], sb, ch,  &(III_side_info->ch[ch].gr[gr]), fr_ps);
    }
    /*Frequency inversion for polyphase.*/
    for (ss=0;ss<18;ss++) 
    {
      for (sb=0; sb<SBLIMIT; sb++)
	if ((ss%2) && (sb%2))
	  hybridOut[sb][ss] = -hybridOut[sb][ss];
    }
    /* Polyphase synthesis */
    for (ss=0;ss<18;ss++)  
    {
      /* 
       ===============================================
       >> BEGIN FIXED-POINT
       ===============================================
      */
      for (sb=0; sb<SBLIMIT; sb++)
	polyPhaseIn_MAD[sb] = mad_f_tofixed( hybridOut[sb][ss] );

      clip += SubBandSynthesis_MAD(polyPhaseIn_MAD, ch, &((*pcm_sample)[ch][ss][0]));
      /* 
       ===============================================
       << END FIXED-POINT
       ===============================================
      */
    }
  }

  /*return clip;*/
  args->clip = clip;
}


/* ============================================================ */
